@@ -21,7 +21,7 @@
   See the License for the specific language governing permissions and
   limitations under the License.
  */
-#include "ts/ink_config.h"
+#include "tscore/ink_config.h"
 #if TS_USE_SET_RBIO
 // No need to do anything, this version of openssl provides the SSL_set0_rbio function
 #else
