@@ -26,11 +26,12 @@ import platform
 import sys
 
 pip_packages = [
-    "autest==1.5.0",
+    "autest==1.7.0",
     "hyper",
     "requests",
     "dnslib",
     "httpbin",
+    "traffic-replay" # this should install TRLib, MicroServer, MicroDNS, Traffic-Replay
 ]
 
 
@@ -48,7 +49,8 @@ distro_packages = {
     "Ubuntu": [
         "install python3",
         "install python3-virtualenv",
-        "install virtualenv"
+        "install virtualenv",
+        "install python3-dev"
     ],
     "CentOS": [
         "install epel-release",
